/*
 *	MP3 huffman table selecting and bit counting
 *
 *	Copyright (c) 1999 Takehiro TOMINAGA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "util.h"
#include "l3side.h"
#include "huffman.h"
#include "quantize.h"


extern long *scalefac_band_long; 
extern long *scalefac_band_short;



struct
{
    unsigned region0_count;
    unsigned region1_count;
} subdv_table[ 23 ] =
{
{0, 0}, /* 0 bands */
{0, 0}, /* 1 bands */
{0, 0}, /* 2 bands */
{0, 0}, /* 3 bands */
{0, 0}, /* 4 bands */
{0, 1}, /* 5 bands */
{1, 1}, /* 6 bands */
{1, 1}, /* 7 bands */
{1, 2}, /* 8 bands */
{2, 2}, /* 9 bands */
{2, 3}, /* 10 bands */
{2, 3}, /* 11 bands */
{3, 4}, /* 12 bands */
{3, 4}, /* 13 bands */
{3, 4}, /* 14 bands */
{4, 5}, /* 15 bands */
{4, 5}, /* 16 bands */
{4, 6}, /* 17 bands */
{5, 6}, /* 18 bands */
{5, 6}, /* 19 bands */
{5, 7}, /* 20 bands */
{6, 7}, /* 21 bands */
{6, 7}, /* 22 bands */
};


/*************************************************************************/
/*	      ix_max							 */
/*************************************************************************/

static long ix_max_takehiro(long *ix, long *end) 
{
    long max = 0;

    while (ix < end) {
	long x =	 *ix++;
	if (max < x) 
	    max = x;
    }
    return max;
}


/*************************************************************************/
/*	      count_bit							 */
/*************************************************************************/

/*
 Function: Count the number of bits necessary to code the subregion. 
*/

static long count_bit_ESC(long *ix, long *end, unsigned long table) 
{
    /* ESC-table is used */
    unsigned char *hlen = ht[table].hlen;
    long linbits = ht[table].linbits;
    long	sum = 0;

    while (ix < end) {
	long x = *ix++;
	long y = *ix++;

	if (x != 0) {
	    sum++;
	    if (x > 14) {
		x = 15;
		sum += linbits;
	    }
	    x *= 16;
	}

	if (y != 0) {
	    sum++;
	    if (y > 14) {
		y = 15;
		sum += linbits;
	    }
	    x += y;
	}

	sum += hlen[x];
    }

    return sum;
}



static long count_bit_noESC(long *ix, long *end, unsigned long table) 
{
    /* No ESC-words */
    long	sum = 0;
    unsigned char *hlen = ht[table].hlen;

    while (ix < end) {
	long x = *ix++;
	long y = *ix++;
	if (x != 0) {
	    sum++;
	    x *= 16;
	}

	if (y != 0) {
	    sum++;
	    x += y;
	}

	sum += hlen[x];
    }

    return sum;
}



static long count_bit_short_ESC(long *ix, long *end, unsigned long table)
{
    /* ESC-table is used */
    unsigned char *hlen = ht[table].hlen;
    unsigned long linbits = ht[table].linbits;
    long	sum = 0;

    while (ix < end) {
	long i;
	for (i = 0; i < 3; i++) {
	    long y = *(ix + 3);
	    long x = *ix++;

	    if (x != 0) {
		sum++;
		if (x > 14) {
		    x = 15;
		    sum += linbits;
		}
		x *= 16;
	    }

	    if (y != 0) {
		sum++;
		if (y > 14) {
		    y = 15;
		    sum += linbits;
		}
		x += y;
	    }

	    sum += hlen[x];
	}
	ix += 3;
    }

    return sum;
}



static long count_bit_short_noESC(long *ix, long *end, unsigned long table) 
{
    /* No ESC-words */
    long	sum = 0;
    unsigned char *hlen = ht[table].hlen;

    while (ix < end) {
	long i;
	for (i = 0; i < 3; i++) {
	    long y = *(ix + 3);
	    long x = *ix++;
	    if (x != 0) {
		sum++;
		x *= 16;
	    }

	    if (y != 0) {
		sum++;
		x += y;
	    }

	    sum += hlen[x];
	}
	ix += 3;
    }

    return sum;
}



/*************************************************************************/
/*	      new_choose table						 */
/*************************************************************************/

/*
  Choose the Huffman table that will encode ix[begin..end] with
  the fewest bits.

  Note: This code contains knowledge about the sizes and characteristics
  of the Huffman tables as defined in the IS (Table B.7), and will not work
  with any arbitrary tables.
*/

long choose_table_takehiro(long *ix, long *end, long * s)
{
    long max;
    long choice0, choice1;
    long sum0, sum1;

    max = ix_max_takehiro(ix, end);

    if (max <= 15)  {
	if (max == 0) {
	    return 0;
	}
	/* try tables with no linbits */
	for (choice0 = 1; choice0 < 14; choice0++) {
	    if (ht[choice0].xlen > max) {
		break;
	    }
	}

	sum0 = count_bit_noESC(ix, end, choice0);
	choice1 = choice0;

	switch (choice0) {
	case 7:
	case 10:
	    choice1++;
	    sum1 = count_bit_noESC(ix, end, choice1);
	    if (sum0 > sum1) {
		sum0 = sum1;
		choice0 = choice1;
	    }
	    /*fall*/
	case 2:
	case 5:
	    choice1++;
	    sum1 = count_bit_noESC(ix, end, choice1);
	    if (sum0 > sum1) {
		sum0 = sum1;
		choice0 = choice1;
	    }
	    break;

	case 13:
	    choice1 += 2;
	    sum1 = count_bit_noESC(ix, end, choice1);
	    if (sum0 > sum1) {
		sum0 = sum1;
		choice0 = choice1;
	    }
	    break;

	default:
	    break;
	}
    } else {
	/* try tables with linbits */
	max -= 15;

	for (choice0 = 15; choice0 < 24; choice0++) {
	    if (ht[choice0].linmax >= max) {
		break;
	    }
	}

	for (choice1 = 24; choice1 < 32; choice1++) {
	    if (ht[choice1].linmax >= max) {
		break;
	    }
	}

	sum0 = count_bit_ESC(ix, end, choice0);
	sum1 = count_bit_ESC(ix, end, choice1);

	if (sum0 > sum1)  {
	    sum0 = sum1;
	    choice0 = choice1;
	}
    }

    *s += sum0;
    return choice0;
}


long choose_table_takehiro_short(long *ix, long *end, long * s)
{
    long max;
    long choice0, sum0;
#define TAKEHIRO_SHORT
#ifdef TAKEHIRO_SHORT
    long choice1, sum1;
#endif

    max = ix_max_takehiro(ix, end);

    if (max <= 15)  {
	if (max == 0) {
	    return 0;
	}
	/* try tables with no linbits */
	for (choice0 = 1; choice0 < 14; choice0++) {
	    if (ht[choice0].xlen > max) {
		break;
	    }
	}

	sum0 = count_bit_short_noESC(ix, end, choice0);

#ifdef TAKEHIRO_SHORT
	choice1 = choice0;

	switch (choice0) {
	case 7:
	case 10:
	    choice1++;
	    sum1 = count_bit_short_noESC(ix, end, choice1);
	    if (sum0 > sum1) {
		sum0 = sum1;
		choice0 = choice1;
	    }
	    /*fall*/
	case 2:
	case 5:
	    choice1++;
	    sum1 = count_bit_short_noESC(ix, end, choice1);
	    if (sum0 > sum1) {
		sum0 = sum1;
		choice0 = choice1;
	    }
	    break;

	case 13:
	    choice1 += 2;
	    sum1 = count_bit_short_noESC(ix, end, choice1);
	    if (sum0 > sum1) {
		sum0 = sum1;
		choice0 = choice1;
	    }
	    break;

	default:
	    break;
	}
#endif
    } else {
	/* try tables with linbits */
	max -= 15;

	for (choice0 = 15; choice0 < 24; choice0++) {
	    if (ht[choice0].linmax >= max) {
		break;
	    }
	}
	sum0 = count_bit_short_ESC(ix, end, choice0);

#ifdef TAKEHIRO_SHORT
	for (choice1 = 24; choice1 < 32; choice1++) {
	    if (ht[choice1].linmax >= max) {
		break;
	    }
	}
	sum1 = count_bit_short_ESC(ix, end, choice1);

	if (sum0 > sum1)  {
	    sum0 = sum1;
	    choice0 = choice1;
	}
#endif
    }

    *s += sum0;
    return choice0;
}


static long count_bits_long(long ix[576], gr_info *gi)
{
    long i, a1, a2;
    long bits = 0;

    i=576;
    for (; i > 1; i -= 2) 
	if (ix[i - 1] | ix[i - 2])
	    break;

    /* Determines the number of bits to encode the quadruples. */
    gi->count1 = i;
    {
	long sum0 = 0, sum1 = 0;
	for (; i > 3; i -= 4) {
	    long p, v;
	    if ((unsigned long)(ix[i-1] | ix[i-2] | ix[i-3] | ix[i-4]) > 1)
		break;

	    v = ix[i-1];
	    p = v;
	    bits += v;

	    v = ix[i-2];
	    if (v != 0) {
		p += 2;
		bits++;
	    }

	    v = ix[i-3];
	    if (v != 0) {
		p += 4;
		bits++;
	    }

	    v = ix[i-4];
	    if (v != 0) {
		p += 8;
		bits++;
	    }

	    sum0 += ht[32].hlen[p];
	    sum1 += ht[33].hlen[p];
	}

	if (sum0 < sum1) {
	    bits += sum0;
	    gi->count1table_select = 0;
	} else {
	    bits += sum1;
	    gi->count1table_select = 1;
	}
    }

    gi->big_values = i;
    if (i == 0)
	return bits;

    if (gi->block_type == 0) {
	long index;
	long scfb_anz = 0;

	while (scalefac_band_long[++scfb_anz] < i) 
	    ;

	index = subdv_table[scfb_anz].region0_count;
	while (scalefac_band_long[index + 1] > i)
	    index--;
	gi->region0_count = index;

	index = subdv_table[scfb_anz].region1_count;
	while (scalefac_band_long[index + gi->region0_count + 2] > i)
	    index--;
	gi->region1_count = index;

	a1 = scalefac_band_long[gi->region0_count + 1];
	a2 = scalefac_band_long[index + gi->region0_count + 2];
	gi->table_select[2] = choose_table_takehiro(ix + a2, ix + i, &bits);
    } else {
	gi->region0_count = 7;
	/*gi->region1_count = SBMAX_l - 7 - 1;*/
	gi->region1_count = SFB_LMAX -1 - 7 - 1;
	a1 = scalefac_band_long[7 + 1];
	a2 = i;
	if (a1 > a2) {
	    a1 = a2;
	}
    }

    /* Count the number of bits necessary to code the bigvalues region. */
    gi->table_select[0] = choose_table_takehiro(ix, ix + a1, &bits);
    gi->table_select[1] = choose_table_takehiro(ix + a1, ix + a2, &bits);
    return bits;
}




long count_bits(long *ix,gr_info *cod_info)  
{
  long bits=0,i;
  for ( i = 0; i < 576; i++ )  {
     if ( ix[i] > 8191 + 14) return 100000;/* report unsuitable quantizer */
  }


  if (cod_info->block_type==2) {
    cod_info->table_select[0] = choose_table_takehiro_short(ix, ix + 36, &bits);
    cod_info->table_select[1] = choose_table_takehiro_short(ix + 36, ix + 576, &bits);
  }else{
    bits=count_bits_long(ix, cod_info);
    cod_info->count1 = (cod_info->count1 - cod_info->big_values) / 4;
    cod_info->big_values /= 2;
  }
  return bits;

}

